local streetterminal = {}


local function GetEdgeObjectSide(segmentEntityId, edgeObjectEntityId)
	local baseEdge = api.engine.getComponent(segmentEntityId, api.type.ComponentType.BASE_EDGE)
	local edgeObjectType = -1
	--commonapi.dmp(baseEdge)
	--print("GetEdgeObjectSide", segmentEntityId, edgeObjectEntityId)
	for m, edgeObject in ipairs(baseEdge.objects) do
		commonapi.dmp(edgeObject)
		if (edgeObject[1] == edgeObjectEntityId) then
			edgeObjectType = edgeObject[2]
		end
	end

	return edgeObjectType
end

local function CalcNumParallelStreetWaypointSignals(segmentEntityId, side)
	local baseEdgeStreet = api.engine.getComponent(segmentEntityId, api.type.ComponentType.BASE_EDGE_STREET)
	local streetType = api.res.streetTypeRep.get(baseEdgeStreet.streetType)
	local laneConfigs = streetType.laneConfigs
	local numSameDir = 0
	if (side == 0) then
		for i = 2, #laneConfigs, 1 do
			if (laneConfigs[i-1].forward == laneConfigs[i].forward) then
				numSameDir = numSameDir + 1
			else
				break
			end
		end
	else
		for i = #laneConfigs-1, 1, -1 do
			if (laneConfigs[i+1].forward == laneConfigs[i].forward) then
				numSameDir = numSameDir + 1
			else
				break
			end
		end
	end
	return numSameDir
end


local function FindEdgeAtInteriorNodeFwd(segmentEntityId, nodeIdx)
	local transportNetwork = api.engine.getComponent(segmentEntityId, api.type.ComponentType.TRANSPORT_NETWORK)

	local edges = transportNetwork.edges
	local tmp = 0
	local fwdIdx = -1
	for i = 1, #edges do
		local edge = edges[i]
		if (edge.conns[1].index == nodeIdx and edge.conns[1].id == segmentEntityId) then
			if ( fwdIdx >= 0 ) then
				print("Logic Error FindEdgeAtInteriorNodeFwd")
			end
			fwdIdx = tmp
		end
		tmp = tmp + 1
	end
	return fwdIdx
end


function streetterminal.getExtraInfoForModelIdOld(modelId)

	local model = api.res.modelRep.get(modelId)
	local description

	if (model.metadata.description ~= nil) then
		description = model.metadata.description.description
	else
		return nil
	end
	local result = {
		vehicleLane = string.match(description, "@@Lane([-]?%d+)"),
		moreCapacity = string.match(description, "@@moreCapacity(%d+)")
	}
	return result
end

function streetterminal.getExtraInfoForModelId(modelId)
	if (commonapi._native == nil) then
		return
	end
	if (commonapi._native.ModelMetadataExt == nil) then
--		print("streetterminal.getExtraInfoForModelId("..modelId..") commonapi._native.ModelMetadataExt not found")
		return
	end

	if (commonapi._native.ModelMetadataExt.isEnabled() ~= true) then
--		print("streetterminal.getExtraInfoForModelId("..modelId..") commonapi._native.ModelMetadataExt.isEnabled ~= true")
		return
	end
--	print("streetterminal.getExtraInfoForModelId("..modelId..")")
	local info = commonapi._native.ModelMetadataExt.get( modelId, "streetTerminalCommonAPI" )
	return info
end

function streetterminal.setStreetNode(params)
	--params.edgeObject
	--commonapi.dmp(params)
	local edgeObjectEntityId = params.edgeObject.edgeObjectEntity
	local segmentEntityId = params.edgeObject.segmentEntity
	local vehicleNodeId = params.vehicleNodeId

	local modelInstance = api.engine.getComponent(edgeObjectEntityId, api.type.ComponentType.MODEL_INSTANCE_LIST)
	local modelId = modelInstance.fatInstances[1].modelId
--	commonapi.dmp(modelInstance)

	local side = GetEdgeObjectSide(segmentEntityId, edgeObjectEntityId)
--	print("side", side)
--	print("modelId ", modelId)

	local model = api.res.modelRep.get(modelId)
	if (model == nil) then
		return
	end
	local info = streetterminal.getExtraInfoForModelId(modelId)
	if (info == nil) then
		info = streetterminal.getExtraInfoForModelIdOld(modelId)
		if (info == nil) then
			return
		end
	end

	local ret = {}
	local hasData = false

	if (info.vehicleLane ~= nil) then
		local laneNum = tonumber(info.vehicleLane)
		local numSameDir = CalcNumParallelStreetWaypointSignals(segmentEntityId, side)
		if (laneNum < 0) then
			laneNum = numSameDir
		end
		if  (numSameDir > 1 and laneNum <= numSameDir) then
			local addDir = laneNum - 1

			ret.vehicleNodeId = {
					id = vehicleNodeId.id,
					index = vehicleNodeId.index + addDir
			}
			hasData = true
		end
	end

	if (info.moreCapacity ~= nil) then
		ret.pool = {
			moreCapacity = tonumber(info.moreCapacity)
		}
		hasData = true
	end
--	local baseEdgeStreet =api.engine.getComponent(segmentEntity, api.type.ComponentType.BASE_EDGE_STREET)
--	commonapi.dmp(baseEdgeStreet)

	if (hasData == true) then
		return ret
	else
		return
	end
end


return streetterminal