local failback = {}

function failback.init()
	if (api and api.gui and api.gui.util and api.gui.comp and api.gui.comp.TextView and commonapi.ui) then
		pcall(failback.initXZ)
	end
end

failback.waitid = "CommonAPI2.MainMduFailbackWait"
function failback.removeWaitId()
	local eleui = api.gui.util.getById(failback.waitid )
	if ( eleui ~= nil) then
		local parent = eleui:getParent()
		local parentLayout = parent:getLayout()
		parentLayout:removeItem(eleui)
		api.gui.util.destroyLater(eleui)
	end
	local menuui = api.gui.util.getById("menuUI")
	if ( menuui ~= nil) then
		local windowcontainer = commonapi.ui._findByName( menuui:getParent() , "WindowContainer", 1)
		if (windowcontainer ~= nil) then
			if (menuui:getLayout():getNumItems() > 4) then
				failback.earlyHooks()
			end
		end
	end
end


function failback.initXZ()
	local waitid = failback.waitid
	if ( api.gui.util.getById(waitid) ) then
		print("Duplicate ID: "..waitid)
		return
	end

	local eleui = api.gui.comp.TextView.new("CommonAPI2 UI waiting for Menu")
	eleui:setId(waitid)
--	local console = api.gui.util.getById("inGameConsole")

--	if (console == nil) then
--		return
--	end

	--local windowcontainer = console:getParent():getParent():getParent()

	local menuui = api.gui.util.getById("menuUI")

	if (menuui == nil) then
		return
	end

	local windowcontainer = commonapi.ui._findByName( menuui:getParent() , "WindowContainer", 1)


	local layout = windowcontainer:getLayout()
	local rect = api.gui.util.Rect.new(10, 10, 100, 100)

	--layout:addItem(eleui, rect)

	local addItemOk
	addItemOk = pcall(layout.addItem, layout, eleui, rect)

	if (addItemOk == false) then
		addItemOk = pcall(layout.addItem, layout, eleui, 0, 0)
		if (addItemOk == false) then
			return
		end
	end

	local dosteps = true
	local stepcounter = 0
	local starttime = os.time()
	local stepfn = function()
		if (dosteps == false) then
			return
		end
		stepcounter = stepcounter + 1
		if (stepcounter < 10) then
			return
		end
		stepcounter = 0
		local curtime = os.time()

		local timediff = (curtime - starttime)
		if ( timediff > 10) then
			--commonapi.dmp("Sec timeout")
			dosteps = false
			eleui:invokeLater(failback.removeWaitId)
			return
		end
		local menui =  api.gui.util.getById("menuUI")
		if ( menui ~= nil) then
			if (menui:getLayout():getNumItems() > 4) then
				eleui:invokeLater(failback.removeWaitId)
			end
		end
	end

	eleui:onStep(stepfn)
end

function failback.earlyHooks()
	failback.showCommonAPISettingsEntry()
end

function failback.getItemFromLayoutByName(layout, name)
	local numitems = layout:getNumItems()
	for i = 1, numitems do
		local child = layout:getItem(i-1)
		if (child ~= nil) then
			if (child.getName ~= nil) then
				if (child:getName() == name) then
					return child
				end
			end
		end
	end
	return nil
end



function failback.clickSettings()
	local ModSettingsWindowTPF = require("commonapi2.uicommon.ModSettingsWindow")
	ModSettingsWindowTPF.createForCommonAPI()
end


function failback.showCommonAPISettingsEntry()
	local buttonId = "CommonAPI2.MainMenuFailbackButton"
	if (api.gui.util.getById(buttonId)) then
		return
	end

	local menuUIele = api.gui.util.getById("menuUI")
	if (menuUIele) then
		local text = "CommonAPI2: "
		local textStyle = ""
		if (commonapi._info.settings.disablenative == true) then
			text = text.._("Native dll/so disabled by settings.lua")
			textStyle = "warning"
		elseif (commonapi._info.versionunknown == true) then
			text = text.."\n".._("Your Transport Fever 2 version ${ugbuild} isn't listed as known good version, native functionality disabled") %
			{ ugbuild = getBuildVersion() }
			textStyle = "negative"
		else
			text = text.. commonapi.version
		end

		local buttontext = api.gui.comp.TextView.new(text)
		local button = api.gui.comp.Button.new(buttontext, true)
		if (textStyle ~= "") then
			buttontext:addStyleClass(textStyle)
			buttontext:addStyleClass("popup")
		end
		button:setId(buttonId)
		--button:setGravity(0, 0.5)


		button:onClick(failback.clickSettings)

		-- menuUIele = api.gui.util.getById("menuUI"); menuuilayout = menuUIele:getLayout()
		local menuuilayout = menuUIele:getLayout()
		menuuilayout:addItem(button, 0, 1)

	end

end
--rawset(_G, "failbackui", failback)

-- api.gui.util.getById("menuUI"):onDestroy(function() print("menuUI Destroy"); end )


local function removeItemFromLayout(layout, oldele)
	if (layout.getNumItems == nil) then
		return false;
	end
	local numitems = layout:getNumItems()
	if (numitems > 0) then
		for i = 1, numitems do
			local child = layout:getItem(i-1)
			if (child == oldele) then
				print("Found layout for "..oldele:getId())
				layout:removeItem(oldele)
				return true
			end
		end
	end
	return false
end


local function doGC()
	if (api and api.gui and api.gui.util) then
		print("GC Called")
		--print("destroying failback ui, found api.gui.util")
		local buttonId = "CommonAPI2.MainMenuFailbackButton"
		local menuUIele = api.gui.util.getById("menuUI")
		local button = api.gui.util.getById(buttonId)
		if (menuUIele ~= nil and button ~= nil) then
			print("Found "..buttonId.." and menuUI")
			local menuuilayout = menuUIele:getLayout()
			if (menuuilayout) then
				print("Found menuUILayout")

				removeItemFromLayout(menuuilayout, button)
				--menuuilayout:removeItem(button)
				print("menuUILayout:removeItem(MainMenuFailbackButton) success")
				if (api.gui.util.destroyLater) then
			--		api.gui.util.destroyLater(button)
				end
				button:destroy()
				print("MainMenuFailbackButton destroy called success")
			end
		end
	end
	print("Destroying failback ui done")
end



function failback.doGC()
	print("GC Called via failback.doGC")
	--doGC()
end

local meta = {
	__gc = function(self)
		doGC()
	end
}
setmetatable(failback, meta)
return failback
